/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021                              ***
***  John Schneiderman <Licensing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_CURRENCY_INTERNAL_VERIFICATION_H_
#define PECUNIA_CURRENCY_INTERNAL_VERIFICATION_H_

#include <cstdint>

#include "pecunia/MoneyTypes.h"


namespace pecunia
{
namespace internal
{

struct FloatingPointParts;

enum class VerifiedValue
{
	Fits,
	Overflows,
	Underflows,
	Undefined
};
class VerifiedFloatingPoint
{
	FloatingPoint value_;

public:
	VerifiedFloatingPoint();
	VerifiedFloatingPoint(const FloatingPoint& value);
	VerifiedFloatingPoint(const std::int32_t& value);
	VerifiedFloatingPoint(const std::int64_t& value);
	VerifiedFloatingPoint(const VerifiedFloatingPoint& vfp) = default;
	VerifiedFloatingPoint(VerifiedFloatingPoint&& vfp) noexcept = default;
	VerifiedFloatingPoint& operator=(const VerifiedFloatingPoint& vfp) = default;
	VerifiedFloatingPoint& operator=( VerifiedFloatingPoint&& vfp) noexcept = default;
	VerifiedFloatingPoint operator+(const VerifiedFloatingPoint& other) const;
	VerifiedFloatingPoint operator-(const VerifiedFloatingPoint& other) const;
	VerifiedFloatingPoint operator*(const VerifiedFloatingPoint& other) const;
	VerifiedFloatingPoint operator/(const VerifiedFloatingPoint& other) const;
	bool operator<(const VerifiedFloatingPoint& other) const;
	bool operator<(const FloatingPoint& other) const;
	bool operator>(const VerifiedFloatingPoint& other) const;
	bool operator>(const FloatingPoint& other) const;
	const FloatingPoint& value() const;
	FloatingPointParts decompose() const;
	VerifiedFloatingPoint floor() const;
	VerifiedFloatingPoint ceil() const;
	VerifiedFloatingPoint abs() const;
	VerifiedFloatingPoint modulo(const VerifiedFloatingPoint& other) const;
	VerifiedFloatingPoint pow(const std::int64_t& exp) const;
	VerifiedFloatingPoint pow(const VerifiedFloatingPoint& exp) const;
	bool isEqual(const VerifiedFloatingPoint& other, const std::uint8_t minorPrecision) const;
	bool isZero(const std::uint8_t minorPrecision) const;
};
struct FloatingPointParts
{
	FloatingPointParts();
	FloatingPointParts(const FloatingPoint& integral, const FloatingPoint& fractional);
	FloatingPointParts(
		const VerifiedFloatingPoint& integral,
		const VerifiedFloatingPoint& fractional
	);
	FloatingPointParts(const FloatingPointParts&) = default;
	FloatingPointParts(FloatingPointParts&&) noexcept = default;
	FloatingPointParts& operator=(const FloatingPointParts&) = default;
	FloatingPointParts& operator=(FloatingPointParts&&) noexcept = default;

	VerifiedFloatingPoint integral_;
	VerifiedFloatingPoint fractional_;
};

// TODO: replace calls with VerifiedFloatingPoint.isEqual.
bool isEqual(
	const FloatingPoint& number,
	const FloatingPoint& equals,
	const std::uint8_t minorPrecision
);
VerifiedValue verifyAdditionFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum
);
VerifiedValue verifySubtractionFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum
);
// TODO: Should use SafeFloatingPoint
VerifiedValue verifyMultiplicationFits(
	const UnitStorage& lhs,
	const FloatingPoint& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum,
	const MinorUnit& minorMaximum,
	const std::uint8_t minorPrecision
);
// TODO: Replace with SafeFloatingPoint
VerifiedValue verifyDivisionFits(
	const UnitStorage& lhs,
	const FloatingPoint& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum,
	const MinorUnit& minorMaximum,
	const std::uint8_t minorPrecision
);
VerifiedValue verifyMultiplicationFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const UnitStorage& maximum,
	const UnitStorage& minimum
);
VerifiedValue verifyDivisionFits(
	const UnitStorage& lhs,
	const UnitStorage& rhs,
	const MinorUnit& minorMaximum
);

}}

#endif
